/* vftext.c -- This is a plug-in for the GIMP (Version 1.0's API).
 * Time-stamp: <1998/05/21 00:45:41 narazaki@InetQ.or.jp>
 * Version: 0.82
 * Author: Shuji Narazaki <narazaki@InetQ.or.jp>
 * URL: http://www.inetq.or.jp/~narazaki/TheGIMP/
 *
 * To run this program, the following program (library) should be installed:
 *
 *  VFlib (c) Hirotsugu KAKUGAWA (kakugawa@se.hiroshima-u.ac.jp)
 *    URL: http://www.se.hiroshima-u.ac.jp/~kakugawa/Hacks/Hacks.html
 *         ftp://gull.se.hiroshima-u.ac.jp/pub/VFlib/
 *
 * These prorgrams are distrbuted under GNU GPL.
 * And this program also must be distributed under the following condition:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Thanks:
 *  Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 *  FUJIWARA Teruyoshi <fujiwara@cim.pe.u-tokyo.ac.jp>
 *  Hide <shide@alles.or.jp>
 *  Tamito Kajiyama <kajiyama@grad.sccs.chukyo-u.ac.jp>
 *
 * Note:
 *  Developement of this program was finished. Only bug-fix.
 */

#include <gtk/gtk.h>
#include "libgimp/gimp.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <VF.h>

/* parameters */
/* vfcap is the name of the VFlib configuration file */
#define VFCAP		NULL
#define MAX_LINE	100	/* how lines vftext can accept?  */
#define LINE_SIZE	1024	/* to hold filename or a line in file */

#define	PLUG_IN_NAME	"plug_in_vftext"
#define SHORT_NAME	"vftext"
#define	MENU_PATH	"<Image>/Filters/Render/VFtext"
#define PROGRESS_NAME	"Rendering Kanji... (version 0.82)"
#define VALS		vfvals
#define INTERFACE	vftext_interface
#define	DIALOG		vftext_dialog
#define ERROR_DIALOG	vftext_error_dialog
#define	OK_CALLBACK	vftext_ok_callback

static void	query	(void);
static void	run	(char	*name,
			 gint	nparams,
			 GParam	*param,
			 gint	*nreturn_vals,
			 GParam **return_vals);
static GStatusType	vftext_render (gint32	image_id,
				       gint32	layer_id,
				       gint32	*new_layer,
				       guchar	*text);
static gint	vftext_parse_configuration_file (gint interactive_p);
static gint	configure_file_parse_line (FILE *file,
					   guchar **true_name,
					   guchar **menu_name);
static guchar	*vftext_get_char_image (gint	jiscode,
					gint	fd,
					gint	xdots,
					gint	ydots,
					gint	rast,
					guchar	*buff);
static gint	vftext_render_to_image (gint32	image_id,
					gint32	*layer_id,
					gint32	beneath_layer,
					gint	fd,
					gint	xdots,
					gint	ydots,
					gint	rast,
					guchar 	*buff,
					guchar	*text);
static void	set_pixel_color (guchar *dest,
				 gint	x,
				 guchar	intensity,
				 gint	ch,
				 guchar	*color);
static void	set_pixel_color_for_preview (guchar *dest,
					     gint	x,
					     guchar	intensity,
					     guchar	*fore,
					     guchar	*back);
static guchar *	vftext_get_string (guchar *file_name);
static gint	vftext_parse_configuration_file (gint interactive_p);
static void	font_table_register (guchar *true_name, guchar *menu_name);
static guchar *	font_table_true_name_of (guchar *name);
static guchar *	vftext_get_path ();
static guchar *	vftext_trim_ASCII (guchar *buffer);
static void	preview_clear ();
static gint	DIALOG ();
static void	vftext_open_fileselector (GtkWidget *widget, gpointer client_data);
static void	vftext_filename_callback (GtkWidget *widget, gpointer data);
static void	ERROR_DIALOG (gint gtk_was_not_initialized, guchar *message);
static void	do_preview_callback (GtkWidget *widget, gpointer   data);
static void	OK_CALLBACK (GtkWidget *widget, gpointer   data);
static void	vftext_open_fontname_dialog (GtkWidget *widget, gpointer data);
static void	vftext_fontname_callback (GtkWidget *widget, gpointer data);

/* gtkWrapper functions */
#define PROGRESS_UPDATE_NUM	100
#define ENTRY_WIDTH	150
#define SCALE_WIDTH	100
#define GTKW_PREVIEW_WIDTH	360
#define GTKW_PREVIEW_HEIGHT	65
#define	GTKW_BORDER_WIDTH	1
gint	gtkW_border_width = GTKW_BORDER_WIDTH;
gint	gtkW_frame_shadow_type = GTK_SHADOW_ETCHED_IN;
gint	gtkW_preview_width = GTKW_PREVIEW_WIDTH;
gint	gtkW_preview_height = GTKW_PREVIEW_HEIGHT;
static void	gtkW_close_callback (GtkWidget *widget, gpointer data);
static void	gtkW_gint_update (GtkWidget *widget, gpointer data);
static void	gtkW_toggle_update (GtkWidget *widget, gpointer data);
static void	gtkW_string_update (GtkWidget *widget, gpointer data);
static GtkWidget * gtkW_error_dialog_new (char * name);
static void gtkW_table_add_toggle (GtkWidget	*table,
				   gchar	*name,
				   gint	x1,
				   gint	x2,
				   gint	y,
				   GtkSignalFunc update,
				   gint	*value);
static void gtkW_table_add_gint (GtkWidget	*table,
				 gchar	*name,
				 gint	x,
				 gint	y,
				 GtkSignalFunc	update,
				 gint	*value,
				 gchar	*buffer);
static GtkWidget *gtkW_frame_new (GtkWidget *parent, gchar *name);
static GtkWidget *gtkW_table_new (GtkWidget *parent, gint col, gint row);
static GtkWidget *gtkW_hbox_new (GtkWidget *parent);
static GtkWidget *gtkW_vbox_new (GtkWidget *parent);
static GtkWidget *gtkW_vbox_add_radio_button (GtkWidget *vbox,
					      gchar	*name,
					      GSList	**group,
					      GtkSignalFunc	update,
					      gint	*value);
static void	gtkW_widget_destroy (GtkWidget *widget, gpointer data);
static void	gtkW_preview_force_to_update (GtkWidget *widget);
static void	gtkW_message_dialog (gint gtk_was_initialized, guchar *message);
static GtkWidget *gtkW_message_dialog_new (guchar *name);
/* gtkW declaration ends here */

typedef struct
{
  guchar	*true_name;
  guchar	*menu_name;
} ftable_entry;

typedef struct
{
  gint		max_index;
  gint		last_index;
  ftable_entry	*font;
} font_table_type;

font_table_type	*font_table = NULL;

GPlugInInfo PLUG_IN_INFO =
{
  NULL,				/* init_proc  */
  NULL,				/* quit_proc */
  query,			/* query_proc */
  run,				/* run_proc */
};

#define READ_AS_STRING	0
#define READ_FROM_FILE	1
typedef struct
{
  guchar	fontname[256];
  gint		fontsize;
  gint		tategaki;
  gint		border;
  gint		spacing;		/* between letters */
  gint		antialiasing;
  gfloat	x;
  gfloat	y;
  gint		string_type;
  guchar	string[LINE_SIZE];
} VFtextValues;

static VFtextValues VALS =
{
  "",			/* font name */
  64,			/* font size */
  0,			/* yokogaki */
  10,			/* border */
  2,			/* spacing */
  1,			/* antialiasing */
  0,			/* x position */
  0,			/* y position */
  READ_AS_STRING,	/* string_kind */
  ""			/* string */
};

typedef struct
{
  gint run;
} VFtextInterface;

static VFtextInterface INTERFACE = { FALSE };

GtkWidget *string_entry = NULL;
GtkWidget *fontname_entry = NULL;
GtkWidget *preview = NULL;
GtkWidget *input_file_button = NULL;
GRunModeType run_mode;

#define A_Ratio      8
/* variables */
static gint	read_from_file = 0;
static gint	read_as_string = 1;
static guchar	*text = NULL;

/* EUC to JIS */
#define KanjiConv(ch1, ch2) (((ch1)&0x7f)*0x100 + ((ch2)&0x7f))

MAIN ();

static void
query ()
{
  static GParamDef args [] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive"},
    { PARAM_IMAGE, "image", "Input image"},
    { PARAM_DRAWABLE, "drawable", "Input drawable (not used)"},
    { PARAM_FLOAT, "x", "the x coordinate for the left side of text bounding box"},
    { PARAM_FLOAT, "y", "the y coordinate for the left side of text bounding box"},
    { PARAM_INT32, "text_type", "if zero, the next parameter is treated as filename. Otherwise it is treated as kanji string itself."},
    { PARAM_STRING, "text", "kanji characters. If this is null string, then text is read from the file that was specified in the last interactive invocation"},
    { PARAM_INT32, "border", "the size of the border: border >= 0" },
    { PARAM_INT32, "antialias", "generate antialiased text" },
    { PARAM_FLOAT, "size", "the size of text in either pixels or points" },
    { PARAM_INT32, "size_type", "not used" },
    { PARAM_STRING, "foundry", "the font foundry, \"*\" for any" },
    { PARAM_STRING, "family", "the font family, \"*\" for any" },
    { PARAM_STRING, "weight", "the font weight, \"*\" for any" },
    { PARAM_STRING, "slant", "the font slant, \"*\" for any" },
    { PARAM_STRING, "set_width", "the font set-width parameter, \"*\" for any" },
    { PARAM_INT32, "spacing", "the font spacing (NOT STRING BUT INTEGER)" }
    };
  static GParamDef return_vals [] =
  {
    { PARAM_DRAWABLE, "text_layer", "ten new kanji text layer"}
  };
  static gint nargs = sizeof (args) / sizeof (args[0]);
  static gint nreturn_vals = sizeof (return_vals) / sizeof (return_vals[0]);;

  gimp_install_procedure (PLUG_IN_NAME,
			  "Add a new layer for kanji characters",
			  "Add a new layer to the image, which holds the kanji letters that are stored in a file (default: ./kanji4gimp.txt) or directly specified kanji string as an argument.",
			  "Shuji Narazaki <narazaki@inetq.or.jp>",
			  "Shuji Narazaki",
			  "1997",
			  MENU_PATH,
			  "RGB*, GRAY*",
			  PROC_PLUG_IN,
			  nargs, nreturn_vals,
			  args, return_vals);
}

static void
run (char	*name,
     gint	nparams,
     GParam	*param,
     gint	*nreturn_vals,
     GParam	**return_vals)
{
  GParam	*values;
  GStatusType	status = STATUS_EXECUTION_ERROR;
  gint32	new_layer;

  values = g_new (GParam, 2);
  run_mode = param[0].data.d_int32;

  *nreturn_vals = 2;
  *return_vals = values;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = status;

  /* 0.82: Call VF_Init only once. [Thu May 21 00:43:20 1998]  */
  if (VF_Init (VFCAP) < 0)
    {
      if (run_mode == RUN_INTERACTIVE)
	{
	  ERROR_DIALOG (FALSE, "Fail to initializing VFlib");
	}
      return;
    }

  if (! vftext_parse_configuration_file ((run_mode != RUN_NONINTERACTIVE)))
    return;

  switch (run_mode)
    {
    case RUN_INTERACTIVE:
      gimp_get_data (PLUG_IN_NAME, &VALS);

      if ((strlen (VALS.fontname) == 0) && font_table)
	strcpy (VALS.fontname, font_table->font[0].menu_name);
      if (VALS.string_type == READ_AS_STRING)
	{
	  read_as_string = TRUE;
	  read_from_file = FALSE;
	}
      else
	{
	  read_as_string = FALSE;
	  read_from_file = TRUE;
	}
      if (! DIALOG ())
	  return;
      if (read_from_file)
	{
	  read_from_file = TRUE;
	  read_as_string = FALSE;
	  VALS.string_type = READ_FROM_FILE;
	}
      else
	{
	  read_as_string = TRUE;
	  read_from_file = FALSE;
	  VALS.string_type = READ_AS_STRING;
	}
      break;
    case RUN_NONINTERACTIVE:
      /* build up the val structure from the parameters */
      VALS.x = param[3].data.d_float;
      VALS.y = param[4].data.d_float;
      VALS.string_type = param[5].data.d_int32;
      text = g_strdup (param[6].data.d_string);
      VALS.border = param[7].data.d_int32;
      VALS.antialiasing = param[8].data.d_int32;
      VALS.fontsize = (gint) param[9].data.d_float;
      /* 10th arg is size_type (not used) */
      strcpy (VALS.fontname, param[11].data.d_string);
      /* 12: family, 13: weight, 14: slant 15: set_width */
      VALS.spacing = param[16].data.d_int32;
      break;
    case RUN_WITH_LAST_VALS:
      gimp_get_data (PLUG_IN_NAME, &VALS);
      break;
    }
  if (VALS.string_type == READ_FROM_FILE)
    text = vftext_get_string (VALS.string);
  else
    text = g_strdup (VALS.string);

  if ((run_mode != RUN_NONINTERACTIVE)
      && ((text == NULL) || (strlen (text) == 0)))
    {
      char	buffer[LINE_SIZE];

      if (VALS.string_type == READ_FROM_FILE) 
	sprintf (buffer, "Error: %s is invalid file name.", VALS.string);
      else
	sprintf (buffer, "Error: empty string");
      ERROR_DIALOG (FALSE, buffer);
      return;
    }

  /* Even if I will fail to render, I must save user's inputs. */
  if (run_mode == RUN_INTERACTIVE)
    gimp_set_data (PLUG_IN_NAME, &VALS, sizeof (VFtextValues));

  status = vftext_render (param[1].data.d_image, param[2].data.d_drawable,
			  &new_layer, text);
  /* 0.82: VF_Deinit should be called only once. [Thu May 21 00:39:18 1998] */
  VF_Deinit ();
  values[0].data.d_status = status;

  if ((VALS.string_type == READ_FROM_FILE) && text)
    g_free (text);
  
  if (run_mode != RUN_NONINTERACTIVE)
    gimp_displays_flush ();

  if (status == STATUS_SUCCESS)
    {
      values[1].type = PARAM_DRAWABLE;
      values[1].data.d_int32 = new_layer;
    }
}

static GStatusType
vftext_render (gint32 image_id,
	       gint32 beneath_layer,
	       gint32 *new_layer,
	       guchar *text)
{
  gint	preview_p;
  gint	fd, xbytes;
  guchar *buff;

  preview_p = (image_id == -1);

  if ((fd = VF_OpenFont (font_table_true_name_of (VALS.fontname))) < 0)
    {
      if (run_mode == RUN_INTERACTIVE)
	{
	  if (preview_p)
	    gtkW_message_dialog (TRUE, "Fail to open for the specified font");
	  else
	    ERROR_DIALOG (FALSE, "Fail to open for the specified font");
	}
      return STATUS_EXECUTION_ERROR;
    }
  xbytes = (VALS.fontsize * A_Ratio + 7) / 8;
  if ((buff = g_malloc (xbytes * VALS.fontsize * A_Ratio * sizeof (char))) == NULL)
    {
      if (run_mode == RUN_INTERACTIVE)
	{
	  if (preview_p)
	    gtkW_message_dialog (TRUE, "g_malloc error");
	  else
	    ERROR_DIALOG (FALSE, "g_malloc error");
	}
      VF_CloseFont (fd);
      return STATUS_EXECUTION_ERROR;
    }
  text = vftext_trim_ASCII (text);
  vftext_render_to_image (image_id, new_layer, beneath_layer, fd,
			  VALS.fontsize, VALS.fontsize,
			  xbytes, buff, text);

  VF_CloseFont (fd);
  g_free (buff);

  return STATUS_SUCCESS;
}

static guchar *
vftext_get_char_image (gint	jiscode,
		       gint	fd,
		       gint	xdots,
		       gint	ydots,
		       gint	rast,
		       guchar	*buff)
{
  memset (buff, (gint) 0, rast * ydots);
  VF_GetBitmap (jiscode, fd, xdots, ydots, rast, 0, buff);
  return buff;
}

static gint
vftext_render_to_image (gint32	image_id,
			gint32	*layer_id,
			gint32	beneath_layer,
			gint	fd,
			gint	xdots,
			gint	ydots,
			gint	rast,
			guchar	*buff,
			guchar	*text)
{
  GPixelRgn	pixel_rgn;
  GDrawable	*drawable = NULL;
  guchar	fore[4];	/* rgb + alpha */
  guchar	back[3];
  guchar	*ylbuff[MAX_LINE];
  guchar	*dest = NULL;
  gint	rendering_width, rendering_height;
  gint	preview_p;
  gint	xshift, yshift, slen, i, x, y, l, n, intensity;
  gint	pixel_count = 0, total_pixel_num;
  gint	index, line_num, slenmax, last_eol;
  gint	len, ch_num;

  preview_p = (image_id == -1);
  for (index = slenmax = line_num = last_eol = 0; index < strlen (text); index++)
    if (text[index] == '\n')
      {
	len = index - last_eol +1;
	if (slenmax < len) slenmax = len;
	ylbuff[line_num] = g_malloc (len);
	memcpy (ylbuff[line_num], text+last_eol, len-1);
	ylbuff[line_num][len-1] = 0; /* newline -> terminator */
	last_eol = index + 1;
	if (preview_p)
	  {
	    line_num++;
	    break;
	  }
	if (MAX_LINE <= ++line_num) break;
      }
  if (last_eol < index)	/* reached eof without newline */
    {
	len = index - last_eol +1;
	if (slenmax < len) slenmax = len;
	ylbuff[line_num] = g_malloc (len);
	memcpy (ylbuff[line_num], text+last_eol, len-1);
	ylbuff[line_num++][len-1] = 0; /* newline -> terminator */
    }

  if (slenmax % 2) slenmax--;
  rendering_width = 2 * VALS.border + slenmax * xdots / 2
    + (slenmax / 2 - 1) * VALS.spacing;
  rendering_height = ydots * line_num + VALS.spacing * (line_num - 1);
  if (! preview_p)
    rendering_height += 2 * VALS.border;

  if ((! preview_p) && VALS.tategaki)
    {
      long temp = rendering_width;
      rendering_width = rendering_height;
      rendering_height = temp;
    }
  total_pixel_num = rendering_width * rendering_height;

  if (preview_p)
    {
      ch_num = 3;

      if (gtkW_preview_height < rendering_height)
	{
	  gtkW_preview_height = rendering_height;
	  gtk_preview_size (GTK_PREVIEW (preview),
			    gtkW_preview_width, gtkW_preview_height);
	  gtk_widget_set_usize (preview,
				gtkW_preview_width, gtkW_preview_height);
	  gtk_widget_show (preview);
	}
    }
  else if (gimp_image_base_type (image_id) == RGB)
    {
      *layer_id = gimp_layer_new (image_id, "kanji text",
				  rendering_width, rendering_height,
				  RGBA_IMAGE, 100, NORMAL_MODE);
      ch_num = 4;		/* rgb + alpha */
    }
  else
    {
      *layer_id = gimp_layer_new (image_id, "kanji text",
				  rendering_width, rendering_height,
				  GRAYA_IMAGE, 100, NORMAL_MODE);
      ch_num = 2;		/* value + alpha */
    }

  if (preview_p)
    {
      preview_clear ();
    }
  else
    {
      gimp_layer_set_preserve_transparency (*layer_id, 0);
      drawable = gimp_drawable_get (*layer_id);

      gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0,
			   rendering_width, rendering_height, TRUE, TRUE);
      /* clear the new layer (make transparent) */
      {
	gpointer	pr;
	gint		x, y;

	pr = gimp_pixel_rgns_register (1, &pixel_rgn);

	for (; pr != NULL; pr = gimp_pixel_rgns_process (pr))
	  {
	    guchar *src = pixel_rgn.data;

	    for (y = 0; y < pixel_rgn.h; y++)
	      {
		guchar *s = src;

		for (x = 0; x < pixel_rgn.w; x++)
		  {
		    gint b;

		    for (b = 0; b < ch_num; b++)
		      *(s + b) = 0;
		    s += pixel_rgn.bpp;
		  }
		src += pixel_rgn.rowstride;
	      }
	  }
      }
    }
  /* use red value if the layer is grayscale */
  gimp_palette_get_foreground (fore+0, fore+1, fore+2);
  gimp_palette_get_background (back+0, back+1, back+2);
  fore[3] = 255;			/* alpha channel */

  if (preview_p)
    {
      xshift = yshift = 1;
    }
  else if (VALS.tategaki)
    {
      xshift = rendering_width - VALS.border - xdots;
      yshift = VALS.border;
    }
  else
    {
      xshift = VALS.border;
      yshift = VALS.border;
    }

  if (! preview_p)
    gimp_progress_init (PROGRESS_NAME);

  dest = (guchar *) g_malloc (rendering_width * ch_num);

  for (n = 0; n < line_num; n++)	/* a line */
    {
      slen = strlen (ylbuff[n]);
      for (i = 0; i < slen -1; i += 2) /* a letter */
	{
	  memset (buff, (gint) 0, rast * ydots);
	  memset (dest, (gint) 0, rendering_width * ch_num);
	  vftext_get_char_image (KanjiConv (ylbuff[n][i], ylbuff[n][i + 1]),
				 fd, xdots*A_Ratio, ydots*A_Ratio, rast, buff);
	  for (y = 0; y < ydots; y++)
	    {
	      for (x = 0; x < xdots; x++)
		{
		  intensity = 0;
		  for (l =0; l < A_Ratio; l++)
		    intensity += buff[y*rast*A_Ratio+l*rast+x];
		  intensity /= A_Ratio;
		  if (preview_p)
		    set_pixel_color_for_preview (dest, x, intensity, fore, back);
		  else
		    set_pixel_color (dest, x, (guchar) intensity, ch_num, fore);
		  if ((! preview_p) && !(++pixel_count % (total_pixel_num / 100)))
		    gimp_progress_update ((double)pixel_count / total_pixel_num);
		}
	      if (preview_p)
		{
		  gint	pw = xdots;
		  gint	py = y;

		  if (gtkW_preview_width < xshift + xdots)
		    pw = gtkW_preview_width - xshift - 1;
		  if (ydots + 1 < gtkW_preview_height)
		    py = y + (gtkW_preview_height - ydots + 1);

		  if (xshift < gtkW_preview_width)
		    gtk_preview_draw_row (GTK_PREVIEW (preview), dest,
					  xshift, py, pw);
		}
	      else
		gimp_pixel_rgn_set_row (&pixel_rgn, dest, xshift, yshift+y, xdots);
	    }
	  if ((! preview_p) && VALS.tategaki)
	    yshift += ydots + VALS.spacing;
	  else
	    xshift += xdots + VALS.spacing;
	}
      if ((! preview_p) && VALS.tategaki)
	{
	  xshift -= xdots + VALS.spacing;
	  yshift = VALS.border;
	}
      else
	{
	  xshift = VALS.border;
	  yshift += ydots + VALS.spacing;
	}
    }
  if (! preview_p)
    {
      gimp_progress_update (1.0);
      gimp_drawable_flush (drawable);
      gimp_layer_set_offsets (drawable->id, VALS.x, VALS.y);
      gimp_drawable_merge_shadow (drawable->id, TRUE);
      gimp_drawable_update (drawable->id, 0, 0,
			    rendering_width, rendering_height);
      gimp_image_add_layer (image_id, *layer_id, -1);
      gimp_drawable_detach (drawable);
    }
  return 1;
}

static void
set_pixel_color (guchar *dest,
		 gint	x,
		 guchar	intensity,
		 gint	ch,
		 guchar	*color)
{
  gint index;
  guchar *base;

  base = dest + x * ch;
  if (! VALS.antialiasing) intensity = (intensity > 127) ? 255 : 0;
  for (index = 0; index < ch; index++)
    *base++ = *color++;
  *(base - 1) = intensity;
}

static void
set_pixel_color_for_preview (guchar *dest,
			     gint	x,
			     guchar	intensity,
			     guchar	*fore,
			     guchar	*back)
{
  gint index;
  guchar *base;

  base = dest + x * 3;
  if (! VALS.antialiasing) intensity = (intensity > 127) ? 255 : 0;
  for (index = 0; index < 3; index++)
    *base++ = ((gint) (*fore++ * intensity) + (gint) (*back++ * (255 - intensity))) / 255.0 ;
}

static guchar *
vftext_get_string (guchar *file_name)
{
  guchar	*m;
  guchar	buff[LINE_SIZE];
  gint	m_len, m_filled;
  FILE	*fin;

#ifdef __EMX__
  if ((fin = fopen (file_name, "rt")) == NULL)
#else
  if ((fin = fopen (file_name, "r")) == NULL)
#endif
    return NULL;
  m = (guchar *) g_malloc (LINE_SIZE);
  m_len = LINE_SIZE;
  m_filled = 0;
  while (fgets (buff, LINE_SIZE - 1, fin))
    {
      if (m_len - m_filled < strlen (buff))
	{
	  m_len += LINE_SIZE;
	  m = (guchar *) g_realloc ((gpointer) m, m_len);
	}
      memcpy (m+m_filled, buff, strlen (buff));
      m_filled += strlen (buff);
    }
  m[m_filled] = 0;
  fclose (fin);
  return m;
}

static gint
vftext_parse_configuration_file (gint interactive_p)
{
  guchar	*filename;
  FILE		*file;

  if ((filename = vftext_get_path (interactive_p)) == NULL)
    return FALSE;

#ifdef __EMX__
  file = fopen (filename, "rt");
#else
  file = fopen (filename, "r");
#endif

  if (! file)
    {
      guchar buffer[LINE_SIZE];

      if (interactive_p)
	{
	  sprintf (buffer, "Error: could not open configuration file:\"%s\"", filename);
	  ERROR_DIALOG (TRUE, buffer);
	}
      return FALSE;
    }
  else
    {
      guchar	*true_name;
      guchar	*menu_name;

      while (configure_file_parse_line (file, &true_name, &menu_name))
	font_table_register (true_name, menu_name);

      return TRUE;
    }
}

static gint
configure_file_parse_line (FILE *file, guchar **true_name, guchar **menu_name)
{
  guchar	line[LINE_SIZE];
  guchar	tmp1[LINE_SIZE];
  guchar	tmp2[LINE_SIZE];

  if (fgets (line, LINE_SIZE -1, file) == NULL)
    return FALSE;
  else
    {
      if ((*line == '\t') || (*line == '#') || (*line == ' '))
	return configure_file_parse_line (file, true_name, menu_name);
      else
	{
	  gint	del1, del2, del3;

	  for (del1 = 0; (line[del1] != '\t') && (line[del1] != '\n'); del1++)
	  if (line[del1] == '\n')
	    return configure_file_parse_line (file, true_name, menu_name);
	  memcpy (tmp1, line, del1);
	  tmp1[del1] = 0;
	  for (del2 = del1; (line[del2] == '\t') || (line[del2] == '\n'); del2++);
	  if (line[del2] == '\n')
	    return configure_file_parse_line (file, true_name, menu_name);
	  for (del3 = del2; (line[del3] != '\t') && (line[del3] != '\n'); del3++);
	  if (del2 == del3)
	    return configure_file_parse_line (file, true_name, menu_name);
	  del3 = del3 - del2;
	  memcpy (tmp2, line + del2, del3);
	  tmp2[del3] = 0;

	  *true_name = (guchar *) g_malloc (del1 + 1);
	  *menu_name = (guchar *) g_malloc (del3 + 1);
	  strcpy (*true_name, tmp1);
	  strcpy (*menu_name, tmp2);
	  return TRUE;
	}
    }
}

static void
font_table_register (guchar *true_name, guchar *menu_name)
{
 /* initialize if needed */
  if (font_table == NULL)
    {
      gint	init_size = 4;

      font_table = (font_table_type *) g_malloc (sizeof (font_table_type));
      font_table->font =
	(ftable_entry *) g_malloc (init_size * sizeof (ftable_entry));
      font_table->max_index = init_size - 1;
      font_table->last_index = -1;
    }

  font_table->last_index++;
 /* expand if needed */
  if (font_table->max_index < font_table->last_index)
    {
      gint	new_size;

      new_size = 2 * (1 + font_table->max_index);
      font_table->font
	= (ftable_entry *) g_realloc ((gpointer) font_table->font,
				      new_size * sizeof (ftable_entry));
      font_table->max_index = new_size - 1;
    }

  font_table->font[font_table->last_index].true_name = true_name;
  font_table->font[font_table->last_index].menu_name = menu_name;
  return;
}

static guchar *
font_table_true_name_of (guchar *name)
{
  gint	index;

  for (index = 0; index <= font_table->last_index; index++)
    if (strcmp (name, font_table->font[index].menu_name) == 0)
      {
	return font_table->font[index].true_name;
      }
  return name;
}

static guchar *
vftext_get_path (gint interactive_p)
{
  GParam	*return_vals;
  gint	nreturn_vals;
  gchar	*path;

  return_vals = gimp_run_procedure ("gimp_gimprc_query",
				    &nreturn_vals,
				    PARAM_STRING, "vftext-configuration-file",
				    PARAM_END);
  if ((return_vals[0].data.d_status != STATUS_SUCCESS)
      || (return_vals[1].data.d_string == NULL))
    {
      if (interactive_p)
	  ERROR_DIALOG (TRUE, "Error: vftext-configuration-file is not defined in your/site .gimprc\nAdd something like:\n (vftext-configratution-file \"~/.gimp/vftext.conf\")");
      gimp_destroy_params (return_vals, nreturn_vals);
      return NULL;
    }
  else
    {
      path = g_strdup (return_vals[1].data.d_string);
      gimp_destroy_params (return_vals, nreturn_vals);
      return path;
    }
}

#ifdef OLD
static guchar *
vftext_trim_ASCII (guchar *buffer)
{
  gint	len, newlen;
  gint	index;
  
  newlen = len = strlen (buffer);
  index = 0;
  
  while (index < newlen)
    {
      if (buffer[index] == ' ')
	{
	  newlen--;
	  memcpy (buffer + index,
		  buffer + (index + 1),
		  newlen - index);
	}
      index++;
    }
  buffer[index] = 0;
  return buffer;
}
#else
static guchar *
vftext_trim_ASCII (guchar *buffer)
{
  GString	*euc;
  guchar	letter[3];
  gint		len;
  gint		index;
  
  euc = g_string_new ("");
  len = strlen (buffer);
  index = 0;
  letter[2] = 0;

  while (index < len)
    {
      if (0x80 & buffer[index])
	{
	  g_string_append_c (euc, buffer[index++]);
	  g_string_append_c (euc, buffer[index++]);
	  continue;
	}
      else if (buffer[index] == '\n')	/* Mon Mar  9 23:26:56 1998 */
	{
	  g_string_append_c (euc, buffer[index++]);
	  continue;
	}
      else if (isspace (buffer[index]))
	strcpy (letter, "");		/* Mon Mar  9 23:26:43 1998 */
      else if (isdigit (buffer[index]))
	{
	  letter[0] = 163;
	  letter[1] = 176 + (buffer[index] -  '0');
	}
      else if (isupper (buffer[index]))
	{
	  letter[0] = 163;
	  letter[1] = 193 + (buffer[index] -  'A');
	}
      else if (islower (buffer[index]))
	{
	  letter[0] = 163;
	  letter[1] = 225 + (buffer[index] -  'a');
	}
      else 		/* How about symbols? I don't know!! */
	{
	  letter[0] = 161;		/* space in iso-2022-jp */
	  letter[1] = 161;
	}

      g_string_append (euc, letter);
      index++;
    }
  if (strlen (buffer) < strlen (euc->str))
    {
      buffer = (guchar *) g_realloc (buffer, strlen (euc->str) + 1);
    }
  strcpy (buffer, euc->str);

  g_string_free (euc, FALSE);
  return buffer;
}
#endif

static void
preview_clear ()
{
  gint	x, y;
  guchar	data[3] = { 255, 255, 255 };

  gimp_palette_get_background (data, data + 1, data + 2);

  for (y = 0; y < gtkW_preview_height; y++)
    for (x = 0; x < gtkW_preview_width; x++)
      gtk_preview_draw_row (GTK_PREVIEW (preview), data, x, y, 1);
  gtk_widget_draw (preview, NULL);
}

/* VFtext interface functions  */
static gint
DIALOG ()
{
  GtkWidget	*dlg;
  GtkWidget	*button;
  GtkWidget	*vbox;
  GtkWidget	*frame;
  GtkWidget	*table;
  GSList	*group = NULL;
  gchar	**argv;
  gint	argc;
  static gchar buffer[3][32];

  argc = 1;
  argv = g_new (gchar *, 1);
  argv[0] = g_strdup (PLUG_IN_NAME);

  gtk_set_locale ();			/* set locale to current locale */
  gtk_init (&argc, &argv);
  gtk_rc_parse (gimp_gtkrc ());

  dlg = gtk_dialog_new ();
  gtk_widget_set_name (dlg, PLUG_IN_NAME);

  button = gtk_button_new_with_label ("Do preview");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) do_preview_callback,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  /* Action Area */
  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dlg)->action_area), 0);
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) OK_CALLBACK,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtkW_close_callback,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  vbox = gtkW_vbox_new (GTK_DIALOG (dlg)->vbox);

  /* file name entry */
  frame = gtkW_frame_new (vbox, "Kanji string");

  {
    GtkWidget	*vbox;
    GtkWidget	*hbox;
    GtkWidget	*button;

    vbox = gtkW_vbox_new (frame);
    gtkW_border_width = 0;
    hbox = gtkW_hbox_new (vbox);
    gtkW_border_width = GTKW_BORDER_WIDTH;
    input_file_button
      = gtkW_vbox_add_radio_button (hbox, "Input filename", &group,
				    (GtkSignalFunc) gtkW_toggle_update,
				    &read_from_file);

    button = gtk_button_new_with_label ("Use file selector");
    gtk_signal_connect (GTK_OBJECT (button),
			"clicked",
			(GtkSignalFunc) vftext_open_fileselector,
			&VALS.string);
    gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 20);
    gtk_widget_show (button);

    gtkW_vbox_add_radio_button (vbox, "Input kanji string", &group,
				(GtkSignalFunc) gtkW_toggle_update,
				&read_as_string);

    string_entry = gtk_entry_new ();
    gtk_widget_set_usize (string_entry, ENTRY_WIDTH, 0);
    gtk_entry_set_text (GTK_ENTRY (string_entry), VALS.string);
    gtk_signal_connect (GTK_OBJECT (string_entry), "changed",
			(GtkSignalFunc) gtkW_string_update,
			&VALS.string);
    gtk_container_add (GTK_CONTAINER (vbox), string_entry);
    gtk_widget_show (string_entry);
  }

  /* parameter settings */
  frame = gtkW_frame_new (vbox, "Parameter Settings");
  table = gtkW_table_new (frame, 7,2); /* 4 raw, 2 columns (name and value) */
  gtk_container_border_width (GTK_CONTAINER (table), 0);

  gtkW_table_add_toggle (table, "Tategaki Mode", 0, 2, 1,
			 (GtkSignalFunc) gtkW_toggle_update,
			 &VALS.tategaki);
  gtkW_table_add_toggle (table, "Antialiasing", 0, 2, 2,
			 (GtkSignalFunc) gtkW_toggle_update,
			 &VALS.antialiasing);
  {
    GtkWidget	*button;

    button = gtk_button_new_with_label ("Font Name");
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			(GtkSignalFunc) vftext_open_fontname_dialog, dlg);
    gtk_table_attach (GTK_TABLE (table), button, 0, 1, 3, 4,
		      GTK_FILL|GTK_EXPAND, GTK_FILL, 5, 0);
    gtk_widget_show (button);

    fontname_entry = gtk_entry_new ();
    gtk_widget_set_usize (GTK_WIDGET (fontname_entry), ENTRY_WIDTH, 0);
    gtk_entry_set_text (GTK_ENTRY (fontname_entry), VALS.fontname);
    gtk_signal_connect (GTK_OBJECT (fontname_entry), "changed",
			(GtkSignalFunc) gtkW_string_update,
			&VALS.fontname);
    gtk_table_attach (GTK_TABLE (table), fontname_entry, 1, 2, 3, 4,
		      GTK_FILL|GTK_EXPAND|GTK_SHRINK, GTK_FILL, 10, 0);
    gtk_widget_show (fontname_entry);
  }

  gtkW_table_add_gint (table, "Font Size", 0, 4,
		       (GtkSignalFunc) gtkW_gint_update,
		       &VALS.fontsize, buffer[0]);
  gtkW_table_add_gint (table, "Border", 0, 5,
		       (GtkSignalFunc) gtkW_gint_update,
		       &VALS.border, buffer[1]);
  gtkW_table_add_gint (table, "Spacing", 0, 6,
		       (GtkSignalFunc) gtkW_gint_update,
		       &VALS.spacing, buffer[2]);
  {
    guchar	*color_cube;
    GtkWidget	*preview_frame;

    gdk_set_use_xshm (gimp_use_xshm ());
    gtk_preview_set_gamma (gimp_gamma ());
    gtk_preview_set_install_cmap (gimp_install_cmap ());
    color_cube = gimp_color_cube ();
    gtk_preview_set_color_cube (color_cube[0], color_cube[1],
				color_cube[2], color_cube[3]);

    gtk_widget_set_default_visual (gtk_preview_get_visual ());
    gtk_widget_set_default_colormap (gtk_preview_get_cmap ());

    preview = gtk_preview_new (GTK_PREVIEW_COLOR);
    gtk_preview_size (GTK_PREVIEW (preview),
		      gtkW_preview_width, gtkW_preview_height);
    preview_frame = gtkW_frame_new (NULL, NULL);
    gtkW_frame_shadow_type = GTK_SHADOW_ETCHED_IN;
    gtk_widget_set_usize (preview_frame, 0, 0);

    gtk_container_border_width (GTK_CONTAINER (preview_frame), 0);
    gtk_container_add (GTK_CONTAINER (preview_frame), preview);
    gtk_box_pack_start (GTK_BOX (vbox), preview_frame, TRUE, TRUE, 0);
    gtk_widget_show (preview);
    preview_clear ();
  }
  gtk_widget_show (dlg);
  gtkW_preview_force_to_update (preview);

  gtk_main ();
  gdk_flush ();

  return INTERFACE.run;
}

static void
vftext_open_fileselector (GtkWidget *widget, gpointer client_data)
{
  GtkWidget *filesel;

  filesel = gtk_file_selection_new ("Read kanji string from");
  gtk_window_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);

  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		      "clicked", (GtkSignalFunc) vftext_filename_callback,
		      filesel);

  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button),
			     "clicked", (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT (GTK_WINDOW (filesel)));
  gtk_widget_show (filesel);

  gtk_main ();
  gdk_flush ();
}

static void
vftext_filename_callback (GtkWidget *widget, gpointer data)
{
  guchar	*pathname, *tmp;

  tmp = gtk_file_selection_get_filename (GTK_FILE_SELECTION (data));
  pathname = (guchar *) g_malloc (strlen (tmp) + 1);
  strcpy (pathname, tmp);
  gtk_widget_destroy (GTK_WIDGET (data));

  strcpy (VALS.string, pathname);

  if (! INTERFACE.run)
    gtk_entry_set_text (GTK_ENTRY (string_entry), VALS.string);
  gtk_widget_draw (string_entry, NULL);

  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (input_file_button),
			       READ_FROM_FILE);
  g_free (pathname);
}

static void
vftext_open_fontname_dialog (GtkWidget *widget, gpointer data)
{
  /* The logo frame & drawing area */
  GtkWidget	*dlg, *vbox, *frame, *button;

  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), "VFtext font selector");
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  /*
    gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc) gtkW_close_callback, NULL);
  */
  /* Action Area */
  button = gtk_button_new_with_label ("Reflect to preview");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) do_preview_callback,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtkW_widget_destroy,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  /*
  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtkW_widget_destroy,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_show (button);
  */
  vbox = gtkW_vbox_new (GTK_DIALOG (dlg)->vbox);
  frame = gtkW_frame_new (vbox, "Font Name");

  {
    GtkWidget	*scrolled_win;
    GtkWidget	*list, *list_item;
    gint	nlist_items = font_table->last_index + 1;
    gint	i;

    scrolled_win = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_container_add (GTK_CONTAINER (frame), scrolled_win);
    list = gtk_list_new ();
    gtk_list_set_selection_mode (GTK_LIST (list), GTK_SELECTION_BROWSE);
    gtk_container_add (GTK_CONTAINER (scrolled_win), list);
    gtk_widget_show (list);

    for (i = 0; i < nlist_items; i++)
      {
	list_item = gtk_list_item_new_with_label (font_table->font[i].menu_name);
	gtk_container_add (GTK_CONTAINER (list), list_item);
	gtk_signal_connect (GTK_OBJECT (list_item), "select",
			    (GtkSignalFunc) vftext_fontname_callback,
			    font_table->font[i].menu_name);
	gtk_widget_show (list_item);
      }
    gtk_widget_set_usize (GTK_WIDGET (scrolled_win), 300, 200);
    gtk_widget_show (scrolled_win);
  }
  gtk_widget_show (dlg);
}

static void
vftext_fontname_callback (GtkWidget *widget, gpointer data)
{
  /*  Is this font being selected?  */
  if (widget->state != GTK_STATE_SELECTED)
    return;

  strcpy (VALS.fontname, data);

  if (! INTERFACE.run)
    gtk_entry_set_text (GTK_ENTRY (fontname_entry), VALS.fontname);
  gtk_widget_draw (fontname_entry, NULL);
}

static void
ERROR_DIALOG (gint gtk_was_not_initialized, guchar *message)
{
  GtkWidget *dlg;
  GtkWidget *table;
  GtkWidget *label;
  gchar	**argv;
  gint	argc;

  if (gtk_was_not_initialized)
    {
      argc = 1;
      argv = g_new (gchar *, 1);
      argv[0] = g_strdup (PLUG_IN_NAME);
      gtk_init (&argc, &argv);
      gtk_rc_parse (gimp_gtkrc ());
    }

  dlg = gtkW_error_dialog_new (PLUG_IN_NAME);

  table = gtk_table_new (1,1, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 10);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), table, TRUE, TRUE, 0);

  label = gtk_label_new (message);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, GTK_FILL|GTK_EXPAND,
		    0, 0, 0);
  gtk_widget_show (label);
  gtk_widget_show (table);

  gtk_main ();
  gdk_flush ();
}

static void
do_preview_callback (GtkWidget *widget,
		     gpointer   data)
{
  if (0 == strlen (VALS.fontname))
    return;

  if (strlen (VALS.string) == 0)
    vftext_render (-1, -1, NULL, g_strdup ("奪"));
  else if (read_as_string)
    vftext_render (-1, -1, NULL, g_strdup (VALS.string));
  else
    {
      guchar *text;

      text = vftext_get_string (VALS.string);
      if ((text == NULL) || (strlen (text) == 0))
	{
	  gtkW_message_dialog (TRUE, "The file does not exist!");
	  vftext_render (-1, -1, NULL, g_strdup ("奪"));
	}
      else
	{
	  vftext_render (-1, -1, NULL, g_strdup (text));
	  g_free (text);
	}
    }
  gtkW_preview_force_to_update (preview);
  gdk_flush ();
}

static void
OK_CALLBACK (GtkWidget *widget,
	      gpointer   data)
{
  INTERFACE.run = TRUE;
  /* gtk_widget_destroy (GTK_WIDGET (data)); */
  gtk_main_quit ();
}

/* VFtext interface functions  */
static void
gtkW_gint_update (GtkWidget *widget,
		  gpointer   data)
{
  *(gint *)data = (gint)atof (gtk_entry_get_text (GTK_ENTRY (widget)));
}

static void
gtkW_close_callback (GtkWidget *widget,
		     gpointer   data)
{
  gtk_main_quit ();
}

static void
gtkW_toggle_update (GtkWidget *widget,
		    gpointer   data)
{
  gint *toggle_val;

  toggle_val = (gint *) data;

  if (GTK_TOGGLE_BUTTON (widget)->active)
    *toggle_val = TRUE;
  else
    *toggle_val = FALSE;
}

/* gtkW is the abbreviation of gtk Wrapper */
static GtkWidget *
gtkW_error_dialog_new (char * name)
{
  GtkWidget *dlg, *button;

  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), name);
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc) gtkW_close_callback, NULL);

  /* Action Area */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      (GtkSignalFunc) gtkW_close_callback, dlg);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);
  gtk_widget_show (dlg);

  return dlg;
}

static GtkWidget *
gtkW_table_new (GtkWidget *parent, gint col, gint row)
{
  GtkWidget	*table;

  table = gtk_table_new (col,row, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 10);
  gtk_container_add (GTK_CONTAINER (parent), table);
  gtk_widget_show (table);

  return table;
}

static GtkWidget *
gtkW_hbox_new (GtkWidget *parent)
{
  GtkWidget	*hbox;

  hbox = gtk_hbox_new (FALSE, gtkW_border_width);
  gtk_container_border_width (GTK_CONTAINER (hbox), gtkW_border_width);
  gtk_box_pack_start (GTK_BOX (parent), hbox, FALSE, TRUE, 0);
  gtk_widget_show (hbox);

  return hbox;
}

static GtkWidget *
gtkW_vbox_new (GtkWidget *parent)
{
  GtkWidget *vbox;

  vbox = gtk_vbox_new (FALSE, gtkW_border_width);
  gtk_container_border_width (GTK_CONTAINER (vbox), gtkW_border_width);
  gtk_container_add (GTK_CONTAINER (parent), vbox);
  gtk_widget_show (vbox);

  return vbox;
}

static GtkWidget *
gtkW_frame_new (GtkWidget *parent,
		gchar *name)
{
  GtkWidget *frame;

  frame = gtk_frame_new (name);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width (GTK_CONTAINER (frame), gtkW_border_width);
  if (parent != NULL)
    gtk_box_pack_start (GTK_BOX (parent), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);

  return frame;
}

static void
gtkW_table_add_toggle (GtkWidget	*table,
		       gchar	*name,
		       gint	x1,
		       gint	x2,
		       gint	y,
		       GtkSignalFunc update,
		       gint	*value)
{
  GtkWidget *toggle;

  toggle = gtk_check_button_new_with_label (name);
  gtk_table_attach (GTK_TABLE (table), toggle, x1, x2, y, y+1,
		    GTK_FILL|GTK_EXPAND, 0, 0, 0);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
		      (GtkSignalFunc) update,
		      value);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle), *value);
  gtk_widget_show (toggle);
}

static void
gtkW_table_add_gint (GtkWidget	*table,
		     gchar	*name,
		     gint	x,
		     gint	y,
		     GtkSignalFunc	update,
		     gint	*value,
		     gchar	*buffer)
{
  GtkWidget *label, *entry;

  label = gtk_label_new (name);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, x, x+1, y, y+1,
		    GTK_FILL|GTK_EXPAND, GTK_FILL, 5, 0);
  gtk_widget_show (label);

  entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), entry, x+1, x+2, y, y+1,
		    GTK_FILL|GTK_EXPAND, GTK_FILL, 10, 0);
  gtk_widget_set_usize (entry, ENTRY_WIDTH, 0);
  sprintf (buffer, "%d", *(gint *)value);
  gtk_entry_set_text (GTK_ENTRY (entry), buffer);
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
		      (GtkSignalFunc) update, value);
  gtk_widget_show (entry);
}

static void
gtkW_string_update (GtkWidget *widget,
		    gpointer   data)
{
  guchar *new_string;

  new_string = gtk_entry_get_text (GTK_ENTRY (widget));
  strcpy (data, new_string);
}

static GtkWidget *
gtkW_vbox_add_radio_button (GtkWidget *vbox,
			    gchar	*name,
			    GSList	**group,
			    GtkSignalFunc	update,
			    gint	*value)
{
  GtkWidget *toggle;

  toggle = gtk_radio_button_new_with_label (*group, name);
  *group = gtk_radio_button_group (GTK_RADIO_BUTTON (toggle));
  gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
		      (GtkSignalFunc) update, value);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle), *value);
  gtk_widget_show (toggle);
  return toggle;
}

static void
gtkW_widget_destroy (GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy (GTK_WIDGET (data));
}

static void
gtkW_preview_force_to_update (GtkWidget *widget)
{
  GtkPreview	*preview = GTK_PREVIEW (widget);
  
  gtk_preview_put (GTK_PREVIEW (widget),
		   widget->window, widget->style->black_gc,
		   (widget->allocation.width - preview->buffer_width) / 2,
		   (widget->allocation.height - preview->buffer_height) / 2,
		   0, 0,
		   widget->allocation.width, widget->allocation.height);
  gdk_flush ();
}

static void
gtkW_message_dialog (gint gtk_was_initialized, guchar *message)
{
  GtkWidget *dlg;
  GtkWidget *table;
  GtkWidget *label;
  gchar	**argv;
  gint	argc;

  if (! gtk_was_initialized)
    {
      argc = 1;
      argv = g_new (gchar *, 1);
      argv[0] = g_strdup (PLUG_IN_NAME);
      gtk_init (&argc, &argv);
      gtk_rc_parse (gimp_gtkrc ());
    }

  dlg = gtkW_message_dialog_new (PLUG_IN_NAME);

  table = gtkW_table_new (GTK_DIALOG (dlg)->vbox, 1, 1);

  label = gtk_label_new (message);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, GTK_FILL|GTK_EXPAND,
		    0, 0, 0);

  gtk_widget_show (label);
  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();
}

static GtkWidget *
gtkW_message_dialog_new (guchar *name)
{
  GtkWidget *dlg, *button;

  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), name);
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      (GtkSignalFunc) gtkW_close_callback, NULL);
  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dlg)->action_area),
			      gtkW_border_width);

  /* Action Area */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  return dlg;
}
/* end of vftext.c */
